#pragma once

#include <pcl\point_types.h>
#include <pcl\point_cloud.h>
#include <pcl\segmentation\sac_segmentation.h>
#include "data_def_api.h"
#include "data_def.h"
#include "logger.h"
#include "imstorage_manager.h"

namespace graft_cv {
	class CRootStockGrabPoint {
	public:
		CRootStockGrabPoint(ConfigParam&c, CGcvLogger*pLog = 0);
		~CRootStockGrabPoint();
		int grab_point_detect(PositionInfo& posinfo);
		float* get_raw_point_cloud(int &data_size);
		int load_data(float*pPoint, int pixel_size, int pt_size, int dtype, const char* fn = 0);
		void set_image_saver(CImStoreManager** ppis) { m_ppImgSaver = ppis; }
	private:
		//global configure parameters
		ConfigParam& m_cparam;
		CGcvLogger * m_pLogger;
		int m_dtype;
		string m_pcdId;

		//ڼ¼һzֵб12ŵ
		float m_1st_row_zmean_rs = -1.0;
		float m_1st_row_zmean_sc = -1.0;

		CImStoreManager** m_ppImgSaver;

		pcl::PointCloud<pcl::PointXYZ>::Ptr m_raw_cloud;
		double m_cloud_mean_dist;

		//ʷλãڼƽλ
		std::vector<int>m_stem_fork_ys;
		int m_stem_fork_ys_size;
		int m_stem_fork_pos_mu;



		int read_ply_file(const char* fn);
		double compute_nearest_neighbor_distance(pcl::PointCloud<pcl::PointXYZ>::Ptr);

		////////////////////////////////////////////////
		//---> 2023-8-8Żseedling orderʶ
		/*void CRootStockGrabPoint::find_tray_top_edge(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			float & tray_top_edge_y
		);
		void CRootStockGrabPoint::find_seedling_position_line_based(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			std::vector<int> &first_seedling_cloud_idx,
			pcl::PointXYZ&xz_center
		);
		void segement_line(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud
		);*/
		//<---------

		/////////////////////////////////////////////////////////////////////////////////////
		//Ҷ޳,ͨŷʽ࣬ȻжÿһǷҶ
		//   ֵ⣺ҶӺ;ͨҶʱ
		void leaf_filter(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud, 
			std::vector<int> &stem_cloud_idx
			);
		//////////////////////////////////////////////////////////////////////////////////////
		//Ҷ޳,̬ͨѧʴõҶ,Ҷڵصĵȥ
		void leaf_filter_morph(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			std::vector<int> &stem_cloud_idx				//output, ˺ĵ
		);
		//3dͼ
		void gen_3d_image(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			pcl::PointXYZ& min_pt,							//input ͼ
			pcl::PointXYZ& max_pt,							//input ͼ
			float step,										//input ͼؼ
			cv::Mat& bin_3d_img,								//output, ɵbinaryͼ
			std::map<int, std::vector<int>>& id2pcdidx
			);
		//3*3*3ڵ
		void dilation_3d(
			cv::Mat& in_3d_img,		//input, ͼ
			cv::Mat& out_3d_img		//output, ɵͼ
		);
		//3*3*3ڵĸʴ
		void erosion_3d(
			cv::Mat& in_3d_img,
			int th,				//ֵ
			cv::Mat& out_3d_img
		);
		//õ
		void get_mass_obj_idx(
			cv::Mat& open_3d_img,				//input, ͼ
			std::map<int, std::vector<int>>& id2pcdidx, //input λõĵ
			std::vector<int>& mass_idx			//output,
		);
		//////////////////////////////////////////////////////////////////////////////////////
		//Ҷ޳,̬ͨѧ
		void leaf_filter_ror(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			std::vector<int> &stem_cloud_idx				//output, ˺ĵ
		);
		void cloud_mean_dist(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			double& mean_dist
		);
		//////////////////////////////////////////////////////////////////////////////////////
		void  CRootStockGrabPoint::find_fork(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,
			float& max_dist,//
			int& max_idx //߽ĵ
		);


		void gen_all_seedling_positions(
			pcl::PointXYZ&key_center,	//룬֪
			std::vector<pcl::PointXYZ>&candidate_center	//б
		);
		bool find_seedling_position_key(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			std::vector<int> &first_seedling_cloud_idx,
			pcl::PointXYZ&xz_center,
			pcl::ModelCoefficients::Ptr& lmodel
		);
		// С
		void nms_box(
			std::vector<pcl::PointXYZ>&clt_root_v,	//ĿĿλ
			std::vector<float>&valid_box_cc_dist,	//Ŀĵĵľ룬ԽԽ
			float hole_step_radius,					//Ŀ뾶
			std::vector<pcl::PointXYZ>& target_toot // ֵ
		);
		//ָͨλ,ȡֵƷǷľ澥λñ浽target_filtered
		void line_filter(pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,  //
			float radius,												//룬ȡ뾶
			float ymin,													//룬yСֵ
			float ymax,													//룬yֵ
			std::vector<pcl::PointXYZ>&target_root,						//룬λ
			std::vector<pcl::PointXYZ>&target_filtered,					//λ
			std::vector<pcl::PointXYZ>&target_filtered_root,			//ϸ
			std::vector<std::vector<int>>&target_filtered_element,		//ϵindex
			std::vector<pcl::ModelCoefficients::Ptr>& target_filtered_models//ֱģ
		);
		void get_line_project_hist(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,	//input	ֱߵ
			pcl::ModelCoefficients::Ptr line_model,				//input		
			std::vector<int>& count_hist					    //  Чֱ1mmڵ
		);

		/////////////////////////////////////////////////
		void find_seedling_position(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			std::vector<int> &first_seedling_cloud_idx,
			pcl::PointXYZ&xz_center
		);
		void crop_nn_analysis(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			pcl::PointCloud<pcl::PointXYZ>::Ptr seed_cloud,
			double dist_mean,
			std::vector<double>&mass_indices,
			std::vector<int>& idx
		);

		void euclidean_clustering_ttsas(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			double d1, double d2,
			std::vector<pcl::PointXYZ>&cluster_center,
			std::vector<std::vector<int>> &clustr_member
		);
		void cal_obb_2d(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			int axis,
			double &dx_obb,
			double &dy_obb,
			double &angle_obb);

		void get_optimal_seed(
			pcl::PointCloud<pcl::PointXYZ>::Ptr,
			pcl::PointXYZ&pt,
			int &pt_idx);
        //֪λãҳпλ
		void tilted_seedling_discover(
			std::vector<pcl::PointXYZ>&cluster_center,	//룬֪
			std::vector<pcl::PointXYZ>&tilted_center	//б
		);

		//ͨȽֱߵƺԭʼͬλxzķΧȷ˵Ƿ޸ŵľ
		/*void get_optimal_seed_compare(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,
			pcl::PointXYZ&pt,
			int &pt_idx,
			std::vector<int>& valid_line_index
		);*/
		//ͨȽֱߵƺԭʼͬλxzķΧȷ˵Ƿ޸ŵľ
		void get_optimal_seed_compare(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,		//input
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,	//input
			pcl::ModelCoefficients::Ptr line_model,			//input
			pcl::PointXYZ&pt,								//
			pcl::PointXYZ &pt_ref,							//	ص㾥ڵĲο
			std::vector<int>& valid_line_index				//
		);
		//һĿռ䷶Χҳֱߣ辥ֱ߷ֲģֱϵĵindex
		void find_line_in_cube(
			pcl::PointCloud<pcl::PointXYZ>::Ptr,	//룬
			pcl::PointXYZ& min_pt_ex,				//룬
			pcl::PointXYZ& max_pt_ex,				//룬
			std::vector<int>& out_idx				//ֱϵindex 
			);

		//ɽͼƬ
		void gen_result_img(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud_raw,//룬δ˵in_cloud_raw, 
			pcl::PointCloud<pcl::PointXYZ>::Ptr,	//룬cloud_dowm_sampled, 
			pcl::PointXYZ& selected_pt,				//룬selected_pt, 
			pcl::PointXYZ& selected_pt_ref,        //룬selected_pt_ref,
			cv::Mat& rst_img						//ͼƬ 640*1280
		);

		void viewer_cloud(pcl::PointCloud<pcl::PointXYZ>::Ptr, std::string&winname);
		void viewer_cloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr, std::string&winname);

		void viewer_cloud_debug(
			pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, 
			pcl::PointXYZ&p,//ץȡ
			pcl::PointXYZ &p_ref,//ֲ
			pcl::PointXYZ &root_pt,
			std::string&winname);

		void viewer_cloud_cluster(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, std::vector<pcl::PointXYZ>cluster_center, std::string&winname);
		void viewer_cloud_cluster_box(
			pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
			std::vector<pcl::PointXYZ>&cluster_center,
			std::vector<pcl::PointXYZ>&cluster_box,
			std::vector<std::vector<int> >& clt_line_idx,
			std::string&winname);
	};

};